@@ -4,11 +4,11 @@ from __future__ import with_statement
 from decimal import Decimal
 from trytond.model import ModelView, ModelSQL, fields, OPERATORS
 from trytond.backend import TableHandler
-from trytond.pyson import In, Eval, Not, In, Equal, If, Get, Bool
+from trytond.pyson import In, Eval, Not, Equal, If, Get, Bool
 from trytond.transaction import Transaction
 
 STATES = {
-    'readonly': In(Eval('state'), ['cancel', 'done']),
+    'readonly': In(Eval('state'), ['cancel', 'assigned', 'done']),
 }
 
 
@@ -35,6 +35,8 @@ class Move(ModelSQL, ModelView):
         on_change_with=['uom']), 'get_unit_digits')
     quantity = fields.Float("Quantity", required=True,
             digits=(16, Eval('unit_digits', 2)), states=STATES)
+    internal_quantity = fields.Float('Internal Quantity', readonly=True,
+        required=True)
     from_location = fields.Many2One("stock.location", "From Location", select=1,
             required=True, states=STATES,
             domain=[('type', 'not in', ('warehouse', 'view'))])
@@ -90,6 +92,9 @@ class Move(ModelSQL, ModelView):
         self._sql_constraints += [
             ('check_move_qty_pos',
                 'CHECK(quantity >= 0.0)', 'Move quantity must be positive'),
+            ('check_move_internal_qty_pos',
+                'CHECK(internal_quantity >= 0.0)',
+                'Internal move quantity must be positive'),
             ('check_from_to_locations',
                 'CHECK(from_location != to_location)',
                 'Source and destination location must be different'),
@@ -108,6 +113,7 @@ class Move(ModelSQL, ModelView):
         ]
         self._constraints += [
             ('check_product_type', 'service_product'),
+            ('check_period_closed', 'period_closed'),
         ]
         self._order[0] = ('id', 'DESC')
         self._error_messages.update({
@@ -116,6 +122,9 @@ class Move(ModelSQL, ModelView):
             'set_state_done': 'You can not set state to done!',
             'del_draft_cancel': 'You can only delete draft or cancelled moves!',
             'service_product': 'You can not use service products for a move!',
+            'period_closed': 'You can not modify move in closed period!',
+            'modify_assigned_done_cancel': ('You can not modify a move '
+                'in the state: "Assigned", "Done" or "Cancel"'),
             })
 
     def init(self, module_name):
@@ -130,8 +139,29 @@ class Move(ModelSQL, ModelView):
                 table.index_action(old_column, action='remove')
             table.drop_fk(old_column)
             table.column_rename(old_column, new_column)
+
+        # Migration from 1.8: new field internal_quantity
+        internal_quantity_exist = table.column_exist('internal_quantity')
+
         super(Move, self).init(module_name)
 
+        # Migration from 1.8: fill new field internal_quantity
+        if not internal_quantity_exist:
+            offset = 0
+            limit = cursor.IN_MAX
+            move_ids = True
+            while move_ids:
+                move_ids = self.search([], offset=offset, limit=limit)
+                offset += limit
+                for move in self.browse(move_ids):
+                    internal_quantity = self._get_internal_quantity(
+                            move.quantity, move.uom, move.product)
+                    self.write(move.id, {
+                        'internal_quantity': internal_quantity,
+                        })
+            table = TableHandler(cursor, self, module_name)
+            table.not_null_action('internal_quantity', action='add')
+
         # Migration from 1.0 check_packing_in_out has been removed
         table  = TableHandler(cursor, self, module_name)
         table.drop_constraint('check_packing_in_out')
@@ -325,6 +355,20 @@ class Move(ModelSQL, ModelView):
                 return False
         return True
 
+    def check_period_closed(self, ids):
+        period_obj = self.pool.get('stock.period')
+        period_ids = period_obj.search([
+            ('state', '=', 'closed'),
+        ], order=[('date', 'DESC')], limit=1)
+        if period_ids:
+            period, = period_obj.browse(period_ids)
+            for move in self.browse(ids):
+                date = (move.effective_date if move.effective_date
+                    else move.planned_date)
+                if date and date < period.date:
+                    return False
+        return True
+
     def get_rec_name(self, ids, name):
         if not ids:
             return {}
@@ -364,7 +408,7 @@ class Move(ModelSQL, ModelView):
                 order=order, count=count, query_string=query_string)
 
     def _update_product_cost_price(self, product_id, quantity, uom, unit_price,
-            currency, company):
+            currency, company, date):
         """
         Update the cost price on the given product
 
@@ -375,6 +419,7 @@ class Move(ModelSQL, ModelView):
         :param unit_price: the unit price of the product
         :param currency: the currency of the unit price
         :param company: the company id ot a BrowseRecord of the company
+        :param date: the date for the currency rate calculation
         """
         uom_obj = self.pool.get('product.uom')
         product_obj = self.pool.get('product.product')
@@ -386,8 +431,6 @@ class Move(ModelSQL, ModelView):
 
         if isinstance(uom, (int, long)):
             uom = uom_obj.browse(uom)
-        if isinstance(currency, (int, long)):
-            currency = currency_obj.browse(currency)
         if isinstance(company, (int, long)):
             company = company_obj.browse(company)
 
@@ -403,8 +446,9 @@ class Move(ModelSQL, ModelView):
         qty = Decimal(str(qty))
         product_qty = Decimal(str(product.template.quantity))
         # convert wrt currency
-        unit_price = currency_obj.compute(currency, unit_price,
-                company.currency, round=False)
+        with Transaction().set_context(date=date):
+            unit_price = currency_obj.compute(currency.id, unit_price,
+                    company.currency.id, round=False)
         # convert wrt to the uom
         unit_price = uom_obj.compute_price(uom, unit_price,
                 product.default_uom)
@@ -427,39 +471,50 @@ class Move(ModelSQL, ModelView):
                 'cost_price': new_cost_price,
                 })
 
+    def _get_internal_quantity(self, quantity, uom, product):
+        uom_obj = self.pool.get('product.uom')
+        internal_quantity = uom_obj.compute_qty(uom, quantity,
+                product.default_uom, round=True)
+        return internal_quantity
+
+
     def create(self, vals):
         location_obj = self.pool.get('stock.location')
         product_obj = self.pool.get('product.product')
+        uom_obj = self.pool.get('product.uom')
         date_obj = self.pool.get('ir.date')
 
+        today = date_obj.today()
         vals = vals.copy()
+        effective_date = vals.get('effective_date') or today
 
+        product = product_obj.browse(vals['product'])
         if vals.get('state') == 'done':
-            if not vals.get('effective_date'):
-                vals['effective_date'] = date_obj.today()
-            currency_id = vals.get('currency', self.default_currency())
-            company_id = vals.get('company', self.default_company())
+            vals['effective_date'] = effective_date
             from_location = location_obj.browse(vals['from_location'])
             to_location = location_obj.browse(vals['to_location'])
-            product = product_obj.browse(vals['product'])
             if from_location.type == 'supplier' \
                     and product.cost_price_method == 'average':
                 self._update_product_cost_price(vals['product'],
                         vals['quantity'], vals['uom'], vals['unit_price'],
-                        currency_id, company_id)
+                        vals['currency'], vals['company'], effective_date)
             if to_location.type == 'supplier' \
                     and product.cost_price_method == 'average':
                 self._update_product_cost_price(vals['product'],
                         -vals['quantity'], vals['uom'], vals['unit_price'],
-                        currency_id, company_id)
+                        vals['currency'], vals['company'], effective_date)
             if not vals.get('cost_price'):
                 # Re-read product to get the updated cost_price
                 product = product_obj.browse(vals['product'])
                 vals['cost_price'] = product.cost_price
 
         elif vals.get('state') == 'assigned':
-            if not vals.get('effective_date'):
-                vals['effective_date'] = date_obj.today()
+            vals['effective_date'] = effective_date
+
+        uom = uom_obj.browse(vals['uom'])
+        internal_quantity = self._get_internal_quantity(vals['quantity'],
+                uom, product)
+        vals['internal_quantity'] = internal_quantity
         return super(Move, self).create(vals)
 
     def write(self, ids, vals):
@@ -467,9 +522,12 @@ class Move(ModelSQL, ModelView):
 
         if isinstance(ids, (int, long)):
             ids = [ids]
+        today = date_obj.today()
+        effective_date = vals.get('effective_date') or today
 
+        moves = self.browse(ids)
         if 'state' in vals:
-            for move in self.browse(ids):
+            for move in moves:
                 if vals['state'] == 'cancel':
                     vals['effective_date'] = False
                     if move.from_location.type == 'supplier' \
@@ -477,13 +535,13 @@ class Move(ModelSQL, ModelView):
                             and move.product.cost_price_method == 'average':
                         self._update_product_cost_price(move.product.id,
                                 -move.quantity, move.uom, move.unit_price,
-                                move.currency, move.company)
+                                move.currency, move.company, today)
                     if move.to_location.type == 'supplier' \
                             and move.state != 'cancel' \
                             and move.product.cost_price_method == 'average':
                         self._update_product_cost_price(move.product.id,
                                 move.quantity, move.uom, move.unit_price,
-                                move.currency, move.company)
+                                move.currency, move.company, today)
 
                 elif vals['state'] == 'draft':
                     if move.state == 'done':
@@ -491,24 +549,37 @@ class Move(ModelSQL, ModelView):
                 elif vals['state'] == 'assigned':
                     if move.state in ('cancel', 'done'):
                         self.raise_user_error('set_state_assigned')
-                    vals['effective_date'] = date_obj.today()
+                    vals['effective_date'] = effective_date
                 elif vals['state'] == 'done':
                     if move.state in ('cancel'):
                         self.raise_user_error('set_state_done')
-                    vals['effective_date'] = date_obj.today()
+                    vals['effective_date'] = effective_date
 
                     if move.from_location.type == 'supplier' \
                             and move.state != 'done' \
                             and move.product.cost_price_method == 'average':
                         self._update_product_cost_price(move.product.id,
                                 move.quantity, move.uom, move.unit_price,
-                                move.currency, move.company)
+                                move.currency, move.company, effective_date)
                     if move.to_location.type == 'supplier' \
                             and move.state != 'done' \
                             and move.product.cost_price_method == 'average':
                         self._update_product_cost_price( move.product.id,
                                 -move.quantity, move.uom, move.unit_price,
-                                move.currency, move.company)
+                                move.currency, move.company, effective_date)
+
+        if reduce(lambda x, y: x or y in vals, ('product', 'uom', 'quantity',
+                'from_location', 'to_location', 'company', 'unit_price',
+                'currency'), False):
+            for move in moves:
+                if move.state in ('assigned', 'done', 'cancel'):
+                    self.raise_user_error('modify_assigned_done_cancel')
+        if reduce(lambda x, y: x or y in vals,
+                ('planned_date', 'effective_date'), False):
+            for move in moves:
+                if move.state in ('done', 'cancel'):
+                    self.raise_user_error('modify_assigned_done_cancel')
+
         res = super(Move, self).write(ids, vals)
 
         if vals.get('state', '') == 'done':
@@ -518,6 +589,15 @@ class Move(ModelSQL, ModelView):
                     self.write(move.id, {
                         'cost_price': move.product.cost_price,
                         })
+
+        for move in self.browse(ids):
+            internal_quantity = self._get_internal_quantity(move.quantity,
+                    move.uom, move.product)
+            if internal_quantity != move.internal_quantity:
+                # Use super to avoid infinite loop
+                super(Move, self).write(move.id, {
+                    'internal_quantity': internal_quantity,
+                    })
         return res
 
     def delete(self, ids):
